import os
import ssl
import asyncio
from io import BytesIO
from shutil import which
from typing import BinaryIO
from urllib.request import getproxies
from tempfile import TemporaryDirectory

from loguru import logger
from lagrange.utils.audio.decoder import decode
from lagrange.utils.httpcat import HttpCat, HttpResponse

try:
    from pysilk import async_encode_file
except ImportError:
    async_encode_file = None


class HttpCatProxies(HttpCat):
    @classmethod
    async def _parse_proxy_response(cls, reader: asyncio.StreamReader) -> HttpResponse:
        stat = await cls._read_line(reader)
        if not stat:
            raise ConnectionResetError
        _, code, status = stat.split(" ", 2)
        header = {}
        cookies = {}
        while True:
            head_block = await cls._read_line(reader)
            if head_block:
                k, v = head_block.split(": ")  # type: str
                if k.title() == "Set-Cookie":
                    name, value = v[: v.find(";")].split("=", 1)
                    cookies[name] = value
                else:
                    header[k.title()] = v
            else:
                break
        return HttpResponse(int(code), status, header, b"", cookies)

    async def connect_http_proxy(self, url: str, conn_timeout=0):
        address, path, with_ssl = self._parse_url(url)
        if conn_timeout:
            reader, writer = await asyncio.wait_for(
                asyncio.open_connection(*address, ssl=with_ssl), conn_timeout
            )
        else:
            reader, writer = await asyncio.open_connection(*address, ssl=with_ssl)
        addr = f"{self.host}:{self.port}"
        await self._request(addr, reader, writer, "CONNECT", addr, header=self.header, wait_rsp=False)
        rsp = await self._parse_proxy_response(reader)

        logger.debug(f"open_tunnel[{rsp.code}]: {rsp.status}")
        if rsp.code == 200:
            self._reader = reader
            self._writer = writer
        else:
            raise ConnectionError(f"proxy error: {rsp.code}")

    async def send_request(
        self, method: str, path: str, body=None, follow_redirect=True, conn_timeout=0
    ) -> HttpResponse:
        if not (self._reader and self._writer):
            proxies = getproxies()
            if "http" in proxies:
                await self.connect_http_proxy(proxies.get("http"), conn_timeout)
                if self.ssl:
                    loop = asyncio.get_running_loop()
                    self._writer._protocol._over_ssl = True  # noqa, suppress warning
                    _transport = await loop.start_tls(
                        self._writer.transport,
                        self._writer.transport.get_protocol(),
                        ssl.create_default_context(),
                        server_side=False,
                        server_hostname=self.host,
                    )
                    self._writer._transport = _transport
        return await super().send_request(method, path, body, follow_redirect, conn_timeout)


async def download_resource(url: str, retry=5, timeout=10) -> bytes:
    if retry > 0:
        try:
            address, path, with_ssl = HttpCatProxies._parse_url(url)
            async with HttpCatProxies(*address, ssl=with_ssl) as req:
                rsp = await req.send_request("GET", path, conn_timeout=timeout)
            length = int(rsp.header.get("Content-Length", 0))
            if length and length != len(rsp.body):
                raise BufferError(f"Content-Length mismatch: {length} != {len(rsp.body)}")
            elif rsp.code != 200:
                raise LookupError(f"Request failed with status {rsp.code}:{rsp.status} {rsp.text()}")
            else:
                return rsp.decompressed_body
        except (asyncio.TimeoutError, BufferError, ConnectionError) as e:
            logger.error(f"Request failed: {repr(e)}")
        except ssl.SSLError as e:
            # Suppress SSL Error
            # like: [SSL: APPLICATION_DATA_AFTER_CLOSE_NOTIFY] application data after close notify (_ssl.c:2706)
            logger.error(f"SSL error: {repr(e)}")
        return await download_resource(url, retry - 1, timeout=timeout)
    else:
        raise ConnectionError(f"Request failed after many tries")


async def transform_audio(audio: BinaryIO) -> BinaryIO:
    try:
        typ = decode(audio)
    except ValueError:
        typ = None

    if typ:
        return audio
    elif not typ and async_encode_file:
        ffmpeg = which("ffmpeg")
        if not ffmpeg:
            raise RuntimeError("ffmpeg not found, transform fail")

        with TemporaryDirectory() as temp_dir:
            input_path = os.path.join(temp_dir, f"{os.urandom(16).hex()}.tmp")
            with open(input_path, "wb") as f:
                f.write(audio.read())

            out_path = os.path.join(temp_dir, f"{os.urandom(16).hex()}.tmp")
            proc = await asyncio.create_subprocess_exec(
                ffmpeg, "-i", input_path, "-f", "s16le", "-ar", "24000", "-ac", "1", "-y", out_path
            )
            if await proc.wait() != 0:
                raise ProcessLookupError(proc.returncode)

            data = await async_encode_file(out_path)
        return BytesIO(data)
    else:
        raise RuntimeError("module 'pysilk-mod' not install, transform fail")



from typing import Dict

uid_dict: Dict[int, str] = {}


def resolve_uid(uin: int) -> str:
    if uin in uid_dict:
        return uid_dict[uin]
    raise ValueError(f"uin {uin} not in uid_dict")


def resolve_uin(uid: str) -> int:
    for k, v in uid_dict.items():
        if v == uid:
            return k
    raise ValueError(f"uid {uid} not found in uid_dict")


def save_uid(uin: int, uid: str) -> None:
    if uin not in uid_dict:
        uid_dict[uin] = uid



import asyncio

from satori import Event
from lagrange.client.client import Client
from lagrange.client.events.friend import FriendMessage
from lagrange.client.events.service import ClientOnline, ClientOffline
from lagrange.client.events.group import (
    GroupRecall,
    GroupMessage,
    GroupMemberQuit,
    GroupNameChanged,
    GroupMemberJoined,
    GroupMemberJoinRequest,
    GroupReaction,
)

from .utils import event_register
from .handler import (
    on_grp_msg,
    on_friend_msg,
    on_grp_recall,
    on_member_quit,
    on_client_online,
    on_member_joined,
    on_client_offline,
    on_grp_name_changed,
    on_grp_member_request,
    on_grp_reaction
)

__all__ = ["apply_event_handler"]
ALL_EVENT_HANDLERS = [
    (GroupMessage, on_grp_msg),
    (FriendMessage, on_friend_msg),
    (GroupRecall, on_grp_recall),
    (ClientOnline, on_client_online),
    (ClientOffline, on_client_offline),
    (GroupMemberJoined, on_member_joined),
    (GroupMemberQuit, on_member_quit),
    (GroupNameChanged, on_grp_name_changed),
    (GroupMemberJoinRequest, on_grp_member_request),
    (GroupReaction, on_grp_reaction)
]


def apply_event_handler(client: Client, queue: asyncio.Queue[Event]):
    for event, ev_handler in ALL_EVENT_HANDLERS:
        event_register(client, queue, event, ev_handler)



from satori import Api
from lagrange.client.client import Client
from satori.server import Adapter, Request

from .types import API_HANDLER


def register_api(
    api: Api,
    adapter: Adapter,
    client: Client,
    handler: API_HANDLER,
):
    @adapter.route(api)
    async def handler_wrapper(request: Request):
        return await handler(client, request)



from __future__ import annotations

import asyncio
from pathlib import Path
from contextlib import suppress
from typing import Set, List, Literal
from datetime import datetime, timedelta

from loguru import logger
from lagrange import version
from qrcode.main import QRCode
from satori.server import Adapter
from lagrange.info import InfoManager
from lagrange.info.app import app_list
from lagrange.client.client import Client
from launart import Launart, any_completed
from satori import User, Login, LoginStatus
from lagrange.utils.sign import sign_provider
from graia.amnesia.builtins.memcache import MemcacheService

from .consts import PLATFORM
from .log import patch_logging
from .apis import apply_api_handlers
from .events import apply_event_handler


class NekoBoxAdapter(Adapter):
    def ensure_manager(self, manager: Launart):
        super().ensure_manager(manager)
        with suppress(ValueError):
            manager.add_component(MemcacheService())

    def get_platform(self) -> str:
        return PLATFORM

    @staticmethod
    def proxy_urls() -> List[str]:
        return [
            "http://thirdqq.qlogo.cn",
            "https://p.qlogo.cn/",
            "https://q1.qlogo.cn",
            "https://gchat.qpic.cn",
            "https://multimedia.nt.qq.com.cn/",
        ]

    async def publisher(self):
        seq = 0
        while True:
            ev = await self.queue.get()
            ev.id = seq
            yield ev
            seq += 1

    def ensure(self, platform: str, self_id: str) -> bool:
        return platform == PLATFORM and self_id == str(self.uin)

    def authenticate(self, token: str) -> bool:
        if token != self.access_token:
            logger.warning("Authentication failed, check upstream token setting.")
            return False
        return True

    async def download_uploaded(self, platform: str, self_id: str, path: str) -> bytes:
        raise NotImplementedError

    async def get_logins(self) -> List[Login]:
        return [
            Login(
                (
                    (LoginStatus.ONLINE if self.client.online.is_set() else LoginStatus.CONNECT)
                    if not self.client._network._stop_flag
                    else LoginStatus.DISCONNECT
                ),  # noqa
                self_id=str(self.client.uin),
                platform=PLATFORM,
                user=User(
                    str(self.client.uin),
                    name=str(self.client.uin),
                    avatar=f"https://q1.qlogo.cn/g?b=qq&nk={self.client.uin}&s=640",
                ),
            )
        ]

    @property
    def required(self) -> Set[str]:
        return set()

    def __init__(
        self,
        uin: int,
        access_token: str,
        sign_url: str | None = None,
        protocol: Literal["linux", "macos", "windows"] = "linux",
        log_level: str = "INFO",
    ):
        self.access_token = access_token
        self.log_level = log_level.upper()

        scope = Path.cwd() / "bots" / str(uin)
        scope.mkdir(exist_ok=True, parents=True)

        self.im = InfoManager(uin, scope / "device.json", scope / "sig.bin")
        self.uin = uin
        self.info = app_list[protocol]
        self.sign = sign_provider(sign_url) if sign_url else None
        self.queue = asyncio.Queue()
        super().__init__()

    client: Client

    @property
    def stages(self) -> Set[Literal["preparing", "blocking", "cleanup"]]:
        return {"preparing", "blocking", "cleanup"}

    async def qrlogin(self, client) -> bool:
        fetch_rsp = await client.fetch_qrcode()
        if isinstance(fetch_rsp, int):
            raise AssertionError(f"Failed to fetch QR code: {fetch_rsp}")
        _, link = fetch_rsp
        logger.debug(link[:-34])
        qr = QRCode()
        qr.add_data(link)
        logger.info("Please use Tencent QQ to scan QR code")
        qr.print_ascii()

        try:
            return await client.qrcode_login(3)
        except AssertionError as e:
            logger.error(f"qrlogin error: {e.args[0]}")
            return False

    async def launch(self, manager: Launart):
        logger.info(f"Running on '{version.__version__}' for {self.uin}")

        with self.im as im:
            if im.sig_info.last_update and (datetime.fromtimestamp(im.sig_info.last_update) + timedelta(30)) < datetime.now():
                logger.warning("siginfo expired")
                im.renew_sig_info()

            self.client = client = Client(
                self.uin,
                self.info,
                im.device,
                im.sig_info,
                self.sign,
            )
            apply_event_handler(client, self.queue)
            apply_api_handlers(self, client)
            async with self.stage("preparing"):
                client.connect()
                success = True
                if im.sig_info.last_update:
                    if (datetime.fromtimestamp(im.sig_info.last_update) + timedelta(14)) > datetime.now():
                        logger.info("try to fast login")
                        if not await client.register():
                            logger.error("fast login failed, try to re-login...")
                            success = await client.easy_login()
                    else:
                        logger.warning("Refresh siginfo")
                        success = await client.easy_login()
                else:
                    success = False

            patch_logging(self.log_level)
            if not success:
                if not await self.qrlogin(client):
                    logger.error("login error")
                else:
                    success = True

            async with self.stage("blocking"):
                if success:
                    im.save_all()
                    await any_completed(manager.status.wait_for_sigexit(), client._network.wait_closed())

            async with self.stage("cleanup"):
                logger.debug("stopping client...")
                await client.stop()

            logger.success("Client stopped")



import sys
import logging
import traceback
from typing import Optional
from types import TracebackType

from loguru import logger
from lagrange.utils.log import log, install_loguru


def loguru_exc_callback(cls: type[BaseException], val: BaseException, tb: Optional[TracebackType], *_, **__):
    """loguru 异常回调

    Args:
        cls (Type[Exception]): 异常类
        val (Exception): 异常的实际值
        tb (TracebackType): 回溯消息
    """
    logger.opt(exception=(cls, val, tb)).error("Exception:")


def loguru_exc_callback_async(loop, context: dict):
    """loguru 异步异常回调

    Args:
        loop (AbstractEventLoop): 异常发生的事件循环
        context (dict): 异常上下文
    """
    message = context.get("message") or "Unhandled exception in event loop"
    if (
        handle := context.get("handle")
    ) and handle._callback.__qualname__ == "ClientConnectionRider.connection_manage.<locals>.<lambda>":
        logger.warning("Uncompleted aiohttp transport", style="yellow bold")
        return
    exception = context.get("exception")
    if exception is None:
        exc_info = False
    else:
        exc_info = (type(exception), exception, exception.__traceback__)
    if (
        "source_traceback" not in context
        and loop._current_handle is not None
        and loop._current_handle._source_traceback
    ):
        context["handle_traceback"] = loop._current_handle._source_traceback

    log_lines = [message]
    for key in sorted(context):
        if key in {"message", "exception"}:
            continue
        value = context[key]
        if key == "handle_traceback":
            tb = "".join(traceback.format_list(value))
            value = "Handle created at (most recent call last):\n" + tb.rstrip()
        elif key == "source_traceback":
            tb = "".join(traceback.format_list(value))
            value = "Object created at (most recent call last):\n" + tb.rstrip()
        else:
            value = repr(value)
        log_lines.append(f"{key}: {value}")

    logger.opt(exception=exc_info).error("\n".join(log_lines))


def patch_logging(level="INFO"):
    for name in logging.root.manager.loggerDict:
        _logger = logging.getLogger(name)
        for handler in _logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                _logger.removeHandler(handler)
    sys.excepthook = loguru_exc_callback
    traceback.print_exception = loguru_exc_callback
    log.set_level(level)
    logger.add(
        "./logs/latest.log",
        format="<g>{time:MM-DD HH:mm:ss}</g> | <lvl>{level: <8}</lvl> | <c><u>{name}</u></c> | <lvl>{message}</lvl>",
        level=level.upper(),
        enqueue=False,
        rotation="00:00",
        compression="zip",
        encoding="utf-8",
        backtrace=True,
        diagnose=True,
        colorize=False,
    )


install_loguru()



import shutil
import asyncio
import secrets
from pathlib import Path
from argparse import ArgumentParser
from configparser import ConfigParser
from typing import List, Optional, overload

from creart import it
from loguru import logger
from satori.server import Server

from nekobox.main import NekoBoxAdapter
from nekobox.log import loguru_exc_callback_async

CONFIG_FILE = Path("nekobox.ini")
cyan = "\033[96m"
reset = "\033[0m"
green = "\033[32m"
gold = "\033[33m"
yellow = "\033[93m"
purple = "\033[95m"
magnet = "\033[35m"
ul = "\033[4m"
bd = "\033[1m"


def run(uin: int, host: str, port: int, token: str, protocol: str, sign_url: str, level: str):
    loop = it(asyncio.AbstractEventLoop)
    loop.set_exception_handler(loguru_exc_callback_async)
    server = Server(host=host, port=port)
    server.apply(NekoBoxAdapter(uin, token, sign_url, protocol, level))  # type: ignore
    server.run()


@overload
def set_cfg(
    cfg: ConfigParser, section: str, option: str, description: str, *, unique: Optional[List[str]] = None
) -> str: ...


@overload
def set_cfg(
    cfg: ConfigParser,
    section: str,
    option: str,
    description: str,
    default: str,
    unique: Optional[List[str]] = None,
) -> str: ...


def set_cfg(
    cfg: ConfigParser,
    section: str,
    option: str,
    description: str,
    default: Optional[str] = None,
    unique: Optional[List[str]] = None,
):
    if not unique:
        unique = []
    try:
        i = str(
            input(f"{description}{f' {cyan}({default}){reset}' if default else f' {cyan}(必填项){reset}'}: ")
        )
        if not i:
            if not default:
                raise ValueError
            else:
                i = default
        if i and unique and i not in unique:
            raise ValueError
        cfg.set(section, option, i)
        return i
    except (TypeError, ValueError):
        print(f">>> 输入不符合约束: {green}{unique}{reset}")
        return set_cfg(cfg, section, option, description, default, unique)  # type: ignore


def generate_cfg(args):
    cfg = ConfigParser()
    if (Path.cwd() / CONFIG_FILE).exists():
        cfg.read(CONFIG_FILE, encoding="utf-8")
    print(f"{cyan}正在生成配置文件...{reset}")
    uin = args.uin
    if not uin:
        uin = set_cfg(cfg, "default", "uin", "Bot 的 QQ 号")
    cfg.add_section(uin)
    set_cfg(cfg, uin, "sign", "Bot 的 SignUrl")
    set_cfg(cfg, uin, "protocol", "Bot 的协议类型", default="linux", unique=["linux", "macos", "windows"])
    set_cfg(cfg, uin, "token", "Satori 的验证 token", default=secrets.token_hex(8))
    set_cfg(cfg, uin, "host", "Satori 服务器绑定地址", default="127.0.0.1")
    set_cfg(cfg, uin, "port", "Satori 服务器绑定端口", default="7777")
    set_cfg(
        cfg,
        uin,
        "log_level",
        "默认日志等级",
        default="INFO",
        unique=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )

    with (Path.cwd() / CONFIG_FILE).open("w+", encoding="utf-8") as f:
        cfg.write(f)
    print(f"{green}配置文件已保存{reset}")


def _run(args):
    if not (Path.cwd() / CONFIG_FILE).exists():
        print(f"请先使用 {yellow}`nekobox gen {args.uin or ''}`{reset} 生成配置文件")
        return
    cfg = ConfigParser()
    cfg.read(CONFIG_FILE, encoding="utf-8")
    uin = args.uin or cfg["default"]["uin"]
    if uin not in cfg:
        print(
            f"账号 {purple}{ul}{uin}{reset} 的相关配置不存在\n请先使用 {yellow}`nekobox gen {uin}`{reset} 生成对应账号的配置文件"
        )
        return
    host = cfg[uin]["host"]
    port = int(cfg[uin]["port"])
    token = cfg[uin]["token"]
    sign_url = cfg[uin]["sign"]
    protocol = cfg[uin]["protocol"]
    level = cfg[uin]["log_level"]
    logger.debug("读取配置文件完成")
    run(int(uin), host, port, token, protocol, sign_url, level)


def _clear(args):
    if (bots := Path("./bots")).exists():
        if not args.uin:
            res = input(
                f"{gold}即将清理: {green}{bots.resolve()} {gold}下的所有数据，是否继续? {bd}{magnet}[y/n] {cyan}(y): {reset}"
            )
            if res.lower() == "y":
                shutil.rmtree(bots)
                print(f"{green}{bots.resolve()}{reset} 清理完毕")
        elif (dir_ := (bots / str(args.uin))).exists():
            res = input(
                f"{gold}即将清理: {green}{dir_.resolve()} {gold}下的所有数据，是否继续? {bd}{magnet}[y/n] {cyan}(y): {reset}"
            )
            if res.lower() == "y":
                shutil.rmtree(dir_)
                print(f"{green}{dir_.resolve()}{reset} 数据清理完毕")
        else:
            print(f"{green}{dir_.resolve()}{reset} 不存在")


def _default(args):
    if not (Path.cwd() / CONFIG_FILE).exists():
        print(f"请先使用 {yellow}`nekobox gen {args.uin or ''}`{reset} 生成配置文件")
        return
    cfg = ConfigParser()
    cfg.read(CONFIG_FILE, encoding="utf-8")
    cfg.set("default", "uin", args.uin)
    with (Path.cwd() / CONFIG_FILE).open("w+", encoding="utf-8") as f:
        cfg.write(f)
    print(f"默认账号已设置为 {purple}{ul}{args.uin}{reset}")


def main():
    parser = ArgumentParser(description=f"{cyan}NekoBox/lagrange-python-satori Server 工具{reset}")
    command = parser.add_subparsers(dest="command", title=f"commands")
    run_parser = command.add_parser("run", help="启动服务器")
    run_parser.add_argument("uin", type=str, nargs="?", help="选择账号")
    run_parser.set_defaults(func=_run)
    gen_parser = command.add_parser("gen", help="生成配置文件")
    gen_parser.add_argument("uin", type=str, nargs="?", help="选择账号")
    gen_parser.set_defaults(func=generate_cfg)
    clean_parser = command.add_parser("clear", help="清除数据")
    clean_parser.add_argument("uin", type=str, nargs="?", help="选择账号")
    clean_parser.set_defaults(func=_clear)
    default_parser = command.add_parser("default", help="设置默认账号")
    default_parser.add_argument("uin", type=str, help="选择账号")
    default_parser.set_defaults(func=_default)
    args = parser.parse_args()

    if args.command:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import sys
import base64
from io import BytesIO
from pathlib import Path
from urllib.parse import quote, unquote
from typing import TYPE_CHECKING, List, Tuple, Union

from loguru import logger
from satori import At as SatoriAt
from satori import Link as SatoriLink
from satori import Text as SatoriText
from satori import Audio as SatoriAudio
from satori import Image as SatoriImage
from satori import Quote as SatoriQuote
from satori import Author as SatoriAuthor
from satori.element import Br as SatoriBr
from satori import Element as SatoriElement
from satori import Message as SatoriMessage
from satori.element import Style as SatoriStyle
from satori.element import Custom as SatoriCustom
from lagrange.client.message.types import Element
from satori.element import Paragraph as SatoriParagraph
from lagrange.client.message.elems import At, Text, AtAll, Audio, Image, Quote, MarketFace

from .utils import transform_audio, download_resource

if TYPE_CHECKING:
    from lagrange.client.client import Client


def encode_data_url(data: Union[str, bytes], mime_type=""):
    if isinstance(data, str):
        encoded = quote(data)
    elif isinstance(data, bytes):
        encoded = base64.b64encode(data)
        mime_type += ";base64"
    else:
        raise TypeError(f"Type {type(data)} not supported")
    return f"data:{mime_type},{encoded}"


def decode_data_url(url: str) -> Tuple[str, bytes]:
    if url.find("data:") != 0:
        raise ValueError("Not a valid Data URL")
    head, data = url[5:].split(",", 1)
    if head.find(";") != -1:
        mime, enc_type = head.split(";", 1)
        if enc_type == "base64":
            decoded = base64.b64decode(data)
        else:
            raise TypeError(f"Type {enc_type} not supported")
    else:
        mime = head
        decoded = unquote(data).encode()
    return mime, decoded


async def parse_resource(url: str) -> bytes:
    logger.debug(f"loading resource: {url[:80]}")
    if url.find("http") == 0:
        return await download_resource(url)
    elif url.find("data") == 0:
        _, data = decode_data_url(url)
        return data
    elif url.find("file://") == 0:
        if sys.platform == "win32":
            path = Path(url[8:])
        else:
            path = Path(url[7:])
        if not path.exists():
            raise FileNotFoundError(f"File not found: {path.absolute()}")
        with open(path, "rb") as f:
            return f.read()
    else:
        raise ValueError("Unsupported URL: %s" % url)


async def msg_to_satori(msgs: List[Element]) -> List[SatoriElement]:
    new_msg: List[SatoriElement] = []
    for m in msgs:
        if isinstance(m, At):
            new_msg.append(SatoriAt(str(m.uin), m.text))
        elif isinstance(m, AtAll):
            new_msg.append(SatoriAt.all())
        elif isinstance(m, Quote):
            new_msg.append(SatoriQuote(str(m.seq))(SatoriAuthor(str(m.uin)), m.msg))
        elif isinstance(m, (Image, MarketFace)):
            new_msg.append(SatoriImage.of(str(m.url), extra={"width": m.width, "height": m.height}))
        elif isinstance(m, Audio):
            new_msg.append(SatoriAudio(m.name))
        elif isinstance(m, Text):
            new_msg.append(SatoriText(m.text))
        else:
            logger.warning("cannot parse message to satori " + repr(m)[:100])
    return new_msg


async def satori_to_msg(client: "Client", msgs: List[SatoriElement], *, grp_id=0, uid="") -> List[Element]:
    new_msg: List[Element] = []
    for m in msgs:
        if isinstance(m, SatoriAt):
            if m.type:
                new_msg.append(AtAll("@全体成员"))
            elif m.id:
                new_msg.append(At(f"@{m.name or m.id}", int(m.id), ""))
        elif isinstance(m, SatoriQuote):
            target = await client.get_grp_msg(grp_id, int(m.id or 0))
            new_msg.append(Quote.build(target[0]))
        elif isinstance(m, SatoriImage):
            data = await parse_resource(m.src)
            if grp_id:
                new_msg.append(await client.upload_grp_image(BytesIO(data), grp_id))
            elif uid:
                new_msg.append(await client.upload_friend_image(BytesIO(data), uid))
            else:
                raise AssertionError
        elif isinstance(m, SatoriText):
            new_msg.append(Text(m.text))
        elif isinstance(m, SatoriLink):
            parsed = await satori_to_msg(client, m._children, grp_id=grp_id, uid=uid)
            new_msg.extend(parsed)
            new_msg.append(Text(f"{': ' if parsed else ''}{m.url}"))
        elif isinstance(m, (SatoriMessage, SatoriStyle)):
            if isinstance(m, SatoriBr):
                new_msg.append(Text("\n"))
            else:
                new_msg.extend(await satori_to_msg(client, m._children, grp_id=grp_id, uid=uid))
                if isinstance(m, SatoriParagraph):
                    new_msg.append(Text("\n"))
        elif isinstance(m, SatoriAudio):
            data = await transform_audio(BytesIO(await parse_resource(m.src)))
            if grp_id:
                new_msg.append(await client.upload_grp_audio(data, grp_id))
            elif uid:
                new_msg.append(await client.upload_friend_audio(data, uid))
            else:
                raise AssertionError
        elif isinstance(m, SatoriCustom):
            if m.type == "template":
                new_msg.extend(
                    await satori_to_msg(client, m._children, grp_id=grp_id, uid=uid)
                )
            else:
                logger.warning("unknown message type on Custom: %s", m.type)
        else:
            logger.warning("cannot trans message to lag " + repr(m)[:100])
    return new_msg



from typing import List, Tuple

from satori.server import Api, Adapter
from lagrange.client.client import Client

from .types import API_HANDLER
from .utils import register_api
from .handler import (
    msg_get,
    login_get,
    msg_create,
    msg_delete,
    channel_list,
    friend_channel,
    guild_get_list,
    guild_member_get,
    guild_member_kick,
    guild_member_list,
    guild_member_mute,
    guild_member_req_approve,
    friend_list,
    reaction_create,
    reaction_delete,
    reaction_clear,
)

__all__ = ["apply_api_handlers"]

ALL_APIS: List[Tuple[Api, API_HANDLER]] = [
    (Api.MESSAGE_CREATE, msg_create),
    (Api.MESSAGE_DELETE, msg_delete),
    (Api.MESSAGE_GET, msg_get),
    (Api.GUILD_MEMBER_KICK, guild_member_kick),
    (Api.GUILD_MEMBER_MUTE, guild_member_mute),
    (Api.GUILD_LIST, guild_get_list),
    (Api.GUILD_MEMBER_GET, guild_member_get),
    (Api.GUILD_MEMBER_LIST, guild_member_list),
    (Api.LOGIN_GET, login_get),
    (Api.CHANNEL_LIST, channel_list),
    (Api.USER_CHANNEL_CREATE, friend_channel),
    (Api.GUILD_MEMBER_APPROVE, guild_member_req_approve),
    (Api.FRIEND_LIST, friend_list),
    (Api.REACTION_CREATE, reaction_create),
    (Api.REACTION_DELETE, reaction_delete),
    (Api.REACTION_CLEAR, reaction_clear)
]


def apply_api_handlers(adapter: Adapter, client: Client):
    for api, api_handler in ALL_APIS:
        register_api(api, adapter, client, api_handler)



from typing import Tuple

# msg type map:
# 1: guild(group)
# 2: private(friend)


def encode_msgid(typ: int, seq: int) -> str:
    if typ == 1:
        msg_id = str(seq)
    elif typ == 2:
        msg_id = f"private:{seq}"
    else:
        raise ValueError(f"Unsupported message type: {typ}")
    return msg_id


def decode_msgid(msg_id: str) -> Tuple[int, int]:
    if msg_id.isdigit():  # guild
        return 1, int(msg_id)
    elif msg_id.find("private:") == 0:
        return 2, int(msg_id[len("private:"):])
    else:
        raise ValueError(f"Invalid msg id: {msg_id}")



from datetime import datetime, timedelta

from loguru import logger
from launart import Launart
from satori.parser import parse
from satori.server import Request, route
from lagrange.client.client import Client
from lagrange.pb.service.group import FetchGrpRspBody
from graia.amnesia.builtins.memcache import MemcacheService
from satori import (
    User,
    Guild,
    Login,
    Member,
    Channel,
    PageResult,
    ChannelType,
    LoginStatus,
    MessageObject,
    transform,
    Role,
)

from ..consts import PLATFORM
from ..uid import resolve_uid
from ..msgid import decode_msgid, encode_msgid
from ..transformer import msg_to_satori, satori_to_msg


async def login_get(client: Client, _request: Request[route.Login]):
    return Login(
        LoginStatus.ONLINE if client.online.is_set() else LoginStatus.OFFLINE,
        self_id=str(client.uin),
        platform=PLATFORM,
        user=User(
            str(client.uin), name=str(client.uin), avatar=f"https://q1.qlogo.cn/g?b=qq&nk={client.uin}&s=640"
        ),
    ).dump()


async def channel_list(client: Client, request: Request[route.ChannelListParam]):
    guild_id = int(request.params["guild_id"])
    guilds = await guild_get_list(client, request)  # type: ignore

    # _next = request.params.get("next")
    guild = next((i for i in guilds.data if i.id == str(guild_id)), None)
    return {
        "data": [Channel(encode_msgid(1, guild_id), ChannelType.TEXT, guild.name if guild else None).dump()]
    }


async def msg_create(client: Client, req: Request[route.MessageParam]):
    typ, uin = decode_msgid(req.params["channel_id"])
    if req.params["content"]:
        tp = transform(
            parse(
                req.params["content"]
            )
        )

        if typ == 1:
            rsp = await client.send_grp_msg(await satori_to_msg(client, tp, grp_id=uin), uin)
        elif typ == 2:
            rsp = await client.send_friend_msg(
                await satori_to_msg(client, tp, uid=resolve_uid(uin)), resolve_uid(uin)
            )
        else:
            raise NotImplementedError(typ)
        return [MessageObject.from_elements(str(rsp), tp)]
    else:
        logger.warning("Empty message, ignore")
        return []


async def msg_delete(client: Client, req: Request[route.MessageOpParam]):
    typ, grp_id = decode_msgid(req.params["channel_id"])
    seq = int(req.params["message_id"])
    if typ == 1:
        rsp = await client.recall_grp_msg(grp_id, seq)
    else:
        raise NotImplementedError(typ)

    return [{"id": str(rsp), "content": "ok"}]


async def msg_get(client: Client, req: Request[route.MessageOpParam]):
    typ, grp_id = decode_msgid(req.params["channel_id"])
    seq = int(req.params["message_id"])
    if typ == 1:
        rsp = (await client.get_grp_msg(grp_id, seq))[0]
    else:
        raise NotImplementedError(typ)

    return MessageObject.from_elements(
        str(rsp),
        await msg_to_satori(rsp.msg_chain),
        channel=Channel(encode_msgid(1, rsp.grp_id), ChannelType.TEXT, rsp.grp_name),
        user=User(str(rsp.uin), rsp.nickname, avatar=f"https://q1.qlogo.cn/g?b=qq&nk={rsp.uin}&s=640"),
    )


async def msg_list(client: Client, req: Request[route.MessageListParam]):
    typ, grp_id = decode_msgid(req.params["channel_id"])
    seq = int(req.params["message_id"])

    if typ == 1:
        rsp = await client.get_grp_msg(grp_id, seq)
    else:
        raise NotImplementedError(typ)

    result = [
        MessageObject.from_elements(
            str(r),
            await msg_to_satori(r.msg_chain),
            channel=Channel(encode_msgid(1, r.grp_id), ChannelType.TEXT, r.grp_name),
            user=User(str(r.uin), r.nickname, avatar=f"https://q1.qlogo.cn/g?b=qq&nk={r.uin}&s=640"),
        ) for r in rsp
    ]


async def guild_member_kick(client: Client, req: Request[route.GuildMemberKickParam]):
    grp_id = int(req.params["guild_id"])
    user_id = int(req.params["user_id"])
    permanent = req.params.get("permanent", False)

    await client.kick_grp_member(grp_id, user_id, permanent)

    return [{"content": "ok"}]


async def guild_member_mute(client: Client, req: Request[route.GuildMemberMuteParam]):
    grp_id = int(req.params["guild_id"])
    user_id = int(req.params["user_id"])
    duration = int(req.params["duration"]) * 1000  # ms to s

    await client.set_mute_member(grp_id, user_id, duration)

    return [{"content": "ok"}]


async def guild_member_list(client: Client, req: Request[route.GuildXXXListParam]):
    grp_id = int(req.params["guild_id"])
    next_key = req.params.get("next")

    rsp = await client.get_grp_members(grp_id, next_key=next_key)

    data = [
        Member(
            user=User(
                id=str(body.account.uin),
                name=body.nickname,
                avatar=f"http://thirdqq.qlogo.cn/headimg_dl?dst_uin={body.account.uin}&spec=640",
            ),
            nick=body.name.string if body.name else body.nickname,
            avatar=f"http://thirdqq.qlogo.cn/headimg_dl?dst_uin={body.account.uin}&spec=640",
            joined_at=datetime.fromtimestamp(body.joined_time),
        ).dump()
        for body in rsp.body
    ]

    return {
        "data": data,
        "next": rsp.next_key,
    }


async def guild_member_get(client: Client, req: Request[route.GuildMemberGetParam]):
    grp_id = int(req.params["guild_id"])
    user_id = int(req.params["user_id"])

    rsp = (await client.get_grp_member_info(grp_id, resolve_uid(user_id))).body[0]

    return [
        Member(
            user=User(
                id=str(rsp.account.uin),
                name=rsp.nickname,
                avatar=f"http://thirdqq.qlogo.cn/headimg_dl?dst_uin={rsp.account.uin}&spec=640",
            ),
            nick=rsp.name.string if rsp.name else rsp.nickname,
            avatar=f"http://thirdqq.qlogo.cn/headimg_dl?dst_uin={rsp.account.uin}&spec=640",
            joined_at=datetime.fromtimestamp(rsp.joined_time),
        ).dump()
    ]


async def guild_get_list(client: Client, req: Request[route.GuildListParam]) -> PageResult[Guild]:
    _next_key = req.params.get("next")
    cache = Launart.current().get_component(MemcacheService).cache

    if data := await cache.get("guild_list"):
        return PageResult(data, None)

    rsp = await client.get_grp_list()
    data = [
        Guild(str(i.grp_id), i.info.grp_name, f"https://p.qlogo.cn/gh/{i.grp_id}/{i.grp_id}/640")
        for i in rsp.grp_list
    ]

    await cache.set("guild_list", data, timedelta(minutes=5))

    return PageResult(data, None)


async def friend_channel(client: Client, req: Request[route.UserChannelCreateParam]):
    user_id = int(req.params["user_id"])
    try:
        pid = resolve_uid(user_id)
    except ValueError:
        pid = req.params.get("guild_id", None)
    return Channel(
        encode_msgid(2, user_id),
        ChannelType.DIRECT,
        str(user_id),
        pid,
    )


async def guild_member_req_approve(client: Client, req: Request[route.ApproveParam]):
    cache = Launart.current().get_component(MemcacheService).cache
    data: FetchGrpRspBody = await cache.get(f"grp_mbr_req#{req.params['message_id']}")
    await client.set_grp_request(
        data.group.grp_id,
        int(req.params["message_id"]),
        data.event_type,
        1 if req.params["approve"] else 2,
        req.params["comment"],
    )
    return [{"content": "ok"}]


async def friend_list(client: Client, req: Request[route.FriendListParam]):
    cache = Launart.current().get_component(MemcacheService).cache

    if data := await cache.get("guild_list"):
        return PageResult(data, None)

    friends = await client.get_friend_list()
    data = [
        User(
            id=str(f.uin),
            name=f.nickname,
            avatar=f"http://thirdqq.qlogo.cn/headimg_dl?dst_uin={f.uin}&spec=640"
        ) for f in friends
    ]

    cache.set("guild_list", data, timedelta(minutes=5))
    return PageResult(data, None)


async def _reaction_process(client: Client, req: Request, is_del: bool):
    typ, grp_id = decode_msgid(req.params["channel_id"])
    seq = int(req.params["message_id"])
    emoji = req.params["emoji"]

    if len(emoji) == 1:
        pass
    elif emoji.find("face:") == 0 and emoji[5:].isdigit():
        emoji = int(emoji[5:])
    else:
        raise ValueError(f"Invalid emoji value '{emoji}'")

    if typ == 1:
        await client.send_grp_reaction(grp_id, seq, emoji, is_cancel=is_del)
    else:
        raise TypeError("Guild only")

    return [{"content": "ok"}]

async def reaction_create(client: Client, req: Request[route.ReactionCreateParam]):
    return await _reaction_process(client, req, False)


async def reaction_delete(client: Client, req: Request[route.ReactionDeleteParam]):
    if "user_id" in req.params and req.params.get("user_id") != client.uin:
        raise ValueError("Cannot delete other user's reaction")
    return await _reaction_process(client, req, True)


async def reaction_clear(client: Client, req: Request[route.ReactionClearParam]):
    return await _reaction_process(client, req, True)



from typing import Optional
from datetime import datetime, timedelta

from loguru import logger
from launart import Launart
from lagrange.client.client import Client
from lagrange.client.events.friend import FriendMessage
from graia.amnesia.builtins.memcache import MemcacheService
from lagrange.client.events.service import ClientOnline, ClientOffline
from satori import (
    User,
    Event,
    Guild,
    Login,
    Member,
    Channel,
    EventType,
    ChannelType,
    LoginStatus,
    MessageObject,
)
from lagrange.client.events.group import (
    GroupRecall,
    GroupMessage,
    GroupMemberQuit,
    GroupNameChanged,
    GroupMemberJoined,
    GroupMemberJoinRequest,
    GroupReaction,
)

from ..consts import PLATFORM
from ..msgid import encode_msgid
from ..transformer import msg_to_satori
from ..uid import save_uid, resolve_uin


async def on_grp_msg(client: Client, event: GroupMessage) -> Optional[Event]:
    save_uid(event.uin, event.uid)
    content = await msg_to_satori(event.msg_chain)
    msg = "".join(str(i) for i in content)
    logger.info(f"{event.grp_name}[{event.nickname}]: {msg!r}")
    usr = User(
        str(event.uin),
        event.nickname,
        avatar=f"https://q1.qlogo.cn/g?b=qq&nk={event.uin}&s=640",
        is_bot=event.is_bot,
    )
    return Event(
        0,
        EventType.MESSAGE_CREATED,
        PLATFORM,
        str(client.uin),
        datetime.fromtimestamp(event.time),
        channel=Channel(encode_msgid(1, event.grp_id), ChannelType.TEXT, event.grp_name),
        guild=Guild(
            str(event.grp_id), event.grp_name, f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        user=usr,
        member=Member(usr, event.nickname, avatar=usr.avatar),
        message=MessageObject.from_elements(str(event.seq), content),
    )


async def on_grp_recall(client: Client, event: GroupRecall) -> Optional[Event]:
    uin = resolve_uin(event.uid)
    usr = User(str(uin), str(uin), avatar=f"https://q1.qlogo.cn/g?b=qq&nk={uin}&s=640")
    return Event(
        0,
        EventType.MESSAGE_DELETED,
        PLATFORM,
        str(client.uin),
        datetime.fromtimestamp(event.time),
        channel=Channel(encode_msgid(1, event.grp_id), ChannelType.TEXT),
        guild=Guild(str(event.grp_id)),
        user=usr,
        member=Member(usr),
        message=MessageObject(str(event.seq), event.suffix),
    )


async def on_friend_msg(client: Client, event: FriendMessage) -> Optional[Event]:
    save_uid(event.from_uin, event.from_uid)
    content = await msg_to_satori(event.msg_chain)
    msg = "".join(str(i) for i in content)
    logger.info(f"{event.from_uin}[{event.from_uid}]: {msg!r}")
    return Event(
        0,
        EventType.MESSAGE_CREATED,
        PLATFORM,
        str(client.uin),
        datetime.fromtimestamp(event.timestamp),
        channel=Channel(encode_msgid(2, event.from_uin), ChannelType.DIRECT, event.from_uid),
        user=User(
            str(event.from_uin), event.from_uid, f"https://q1.qlogo.cn/g?b=qq&nk={event.from_uin}&s=640"
        ),
        message=MessageObject.from_elements(str(event.seq), content),
    )


async def on_client_online(client: Client, event: ClientOnline) -> Optional[Event]:
    logger.debug("login-updated: online")
    return Event(
        0,
        EventType.LOGIN_UPDATED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        login=Login(
            LoginStatus.ONLINE,
            self_id=str(client.uin),
            platform=PLATFORM,
            user=User(
                str(client.uin),
                name=str(client.uin),
                avatar=f"https://q1.qlogo.cn/g?b=qq&nk={client.uin}&s=640",
            ),
        ),
    )


async def on_client_offline(client: Client, event: ClientOffline) -> Optional[Event]:
    logger.debug(f"login-updated: {'reconnect' if event.recoverable else 'disconnect'}")
    return Event(
        0,
        EventType.LOGIN_UPDATED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        login=Login(
            LoginStatus.RECONNECT if event.recoverable else LoginStatus.DISCONNECT,
            self_id=str(client.uin),
            platform=PLATFORM,
            user=User(
                str(client.uin),
                name=str(client.uin),
                avatar=f"https://q1.qlogo.cn/g?b=qq&nk={client.uin}&s=640",
            ),
        ),
    )


async def on_grp_name_changed(client: Client, event: GroupNameChanged) -> Event:
    operator_id = resolve_uin(event.operator_uid)
    return Event(
        0,
        EventType.GUILD_UPDATED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        guild=Guild(
            str(event.grp_id), event.name_new, f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        operator=User(
            str(operator_id), str(operator_id), f"https://q1.qlogo.cn/g?b=qq&nk={operator_id}&s=640"
        ),
    )


async def on_member_joined(client: Client, event: GroupMemberJoined) -> Event:
    rsp = (await client.get_grp_member_info(event.grp_id, event.uid)).body[0]
    nickname = rsp.name.string if rsp.name else None
    user = User(
        str(event.uin), str(rsp.nickname), nickname, avatar=f"https://q1.qlogo.cn/g?b=qq&nk={event.uin}&s=640"
    )
    return Event(
        0,
        EventType.GUILD_MEMBER_ADDED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        guild=Guild(
            str(event.grp_id), str(event.grp_id), f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        member=Member(
            user,
            user.nick or user.name,
            avatar=user.avatar,
            joined_at=datetime.fromtimestamp(rsp.joined_time),
        ),
        user=user,
    )


async def on_member_quit(client: Client, event: GroupMemberQuit) -> Optional[Event]:
    user = User(str(event.uin), str(event.uin), avatar=f"https://q1.qlogo.cn/g?b=qq&nk={event.uin}&s=640")
    operator = None
    if event.is_kicked and event.operator_uid:
        operator_id = resolve_uin(event.operator_uid)
        operator = User(
            str(operator_id), str(operator_id), avatar=f"https://q1.qlogo.cn/g?b=qq&nk={operator_id}&s=640"
        )
    return Event(
        0,
        EventType.GUILD_MEMBER_REMOVED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        guild=Guild(
            str(event.grp_id), str(event.grp_id), f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        member=Member(user, user.nick or user.name, avatar=user.avatar),
        user=user,
        operator=operator,
    )


async def on_grp_member_request(client: Client, event: GroupMemberJoinRequest) -> Optional[Event]:
    reqs = await client.fetch_grp_request(4)
    for req in reqs.requests:
        if req.group.grp_id == event.grp_id and req.target.uid == event.uid:
            break
    else:
        return
    cache = Launart.current().get_component(MemcacheService).cache
    await cache.set(f"grp_mbr_req#{req.seq}", req, timedelta(minutes=30))
    user_id = resolve_uin(event.uid)
    user = User(str(user_id), req.target.name, avatar=f"https://q1.qlogo.cn/g?b=qq&nk={user_id}&s=640")
    return Event(
        0,
        EventType.GUILD_MEMBER_REQUEST,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        guild=Guild(
            str(event.grp_id), req.group.grp_name, f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        user=user,
        member=Member(user),
        message=MessageObject(id=str(req.seq), content=req.comment),
    )


async def on_grp_reaction(client: Client, event: GroupReaction) -> Optional[Event]:
    user_id = resolve_uin(event.uid)

    if event.is_emoji:
        emoji = chr(event.emoji_id)
    else:
        emoji = f"face:{event.emoji_id}"

    return Event(
        0,
        EventType.REACTION_ADDED if event.is_increase else EventType.REACTION_REMOVED,
        PLATFORM,
        str(client.uin),
        datetime.now(),
        guild=Guild(
            str(event.grp_id), str(event.grp_id), avatar=f"https://p.qlogo.cn/gh/{event.grp_id}/{event.grp_id}/640"
        ),
        user=User(str(user_id), str(user_id), avatar=f"https://q1.qlogo.cn/g?b=qq&nk={user_id}&s=640"),
        _type="reaction",
        _data={"message_id": event.seq, "emoji": emoji, "count": event.emoji_count}
    )



import asyncio
from typing import Type, TypeVar, Callable, Optional, Coroutine

from loguru import logger
from satori import EventType
from satori.server import Event
from lagrange.client.client import Client
from lagrange.client.events import BaseEvent

TEvent = TypeVar("TEvent", bound=BaseEvent)


def event_register(
    client: Client,
    queue: asyncio.Queue[Event],
    event_type: Type[TEvent],
    handler: Callable[["Client", TEvent], Coroutine[None, None, Optional[Event]]],
):
    async def _after_handle(_client: Client, event: TEvent):
        ev = await handler(_client, event)
        if ev:
            if ev.type != EventType.MESSAGE_CREATED:
                logger.debug(f"Event '{ev.type}' was triggered")
            await queue.put(ev)

    client.events.subscribe(event_type, handler=_after_handle)
